import datetime
import json
import queue
import threading
import time

import requests
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

from .logger import logger
from .request import StReq
from .session import HttpTransaction
from .session_maintainer import SessionMaintainerBase


# Client 用于收发HTTP请求的
class Client:
    http_session_lock = threading.Lock()
    http_session_queue = queue.Queue()

    def __init__(self, session, session_maintainer: SessionMaintainerBase):
        self.session = session
        self.session_maintainer = session_maintainer
        self.http_session = self.get_http_session()

    # 这四个函数从session中拿信息做处理
    def run(self):
        if self.session_maintainer.init_session is not None:
            self.session_maintainer.init_session(self.session)

        while True:
            http_trans = HttpTransaction(
                url=self.session_maintainer.url,
                method=self.session_maintainer.http_method,
                request=None, response=None, status_code=None,
                request_time=None)

            req = self.session_maintainer.wrap_req(self.session)
            if not isinstance(req, StReq):
                req = StReq(req)
            if req.url is None:
                req.url = self.session_maintainer.url
            if req.http_method is None:
                req.http_method = self.session_maintainer.http_method
            if req.headers is None:
                req.headers = {}
            if isinstance(req.req_data, (dict, list)):
                req.req_data = json.JSONEncoder().encode(req.req_data)
                req.headers["Content-Type"] = "application/json"

            def send_request():
                http_trans.request = req.req_data
                http_trans.request_time = datetime.datetime.now()
                if req.http_method == "GET":
                    r_ = self.http_session.get(req.url, params=req.req_data, headers=req.headers, timeout=req.timeout)
                elif req.http_method == "POST":
                    r_ = self.http_session.post(req.url, data=req.req_data, headers=req.headers, timeout=req.timeout)
                else:
                    raise RuntimeError(f"unsupported http method: {req.http_method}")
                end_time = datetime.datetime.now()  # 记录结束时间
                elapsed_time = (end_time - http_trans.request_time).total_seconds()  # 计算请求时间
                return r_, elapsed_time

            r = None
            cost = 0
            for _ in range(req.retry + 1):
                try:
                    r, cost = send_request()
                    if r.status_code == 200:
                        break
                except:
                    time.sleep(0.5)
                http_trans.retry_cnt += 1

            if r is None:
                self.session.append_transaction(http_trans)
                logger.error(f"break session, failed to send request: {req}")
                break
            http_trans.status_code = r.status_code
            http_trans.response = r.text
            http_trans.cost_time = cost
            self.session.append_transaction(http_trans)
            if r.status_code != 200:
                logger.error("break session, "
                             f"failed to send request: {req}, status_cod: {r.status_code}, rsp: {r.text}")
                break

            if self.session_maintainer.update_session is not None:
                self.session_maintainer.update_session(self.session)

            if self.session_maintainer.should_stop_session is None or \
                    self.session_maintainer.should_stop_session(self.session):
                break

    def __del__(self):
        self.release_session(self.http_session)

    @classmethod
    def get_http_session(cls):
        with cls.http_session_lock:
            try:
                return cls.http_session_queue.get_nowait()
            except:
                s = requests.Session()

                # 定义重试策略
                retry_strategy = Retry(
                    total=3,  # 总共重试次数
                    backoff_factor=1,  # 重试间隔时间的倍数
                    status_forcelist=[429, 500, 502, 503, 504],  # 需要重试的HTTP状态码
                    allowed_methods=["HEAD", "GET", "OPTIONS", "POST"]  # 需要重试的方法
                )

                # 创建一个适配器并将其安装到会话中
                adapter = HTTPAdapter(max_retries=retry_strategy)
                s.mount("http://", adapter)
                s.mount("https://", adapter)
                return s

    @classmethod
    def release_session(cls, http_session):
        with cls.http_session_lock:
            cls.http_session_queue.put(http_session)



from dataclasses import dataclass, asdict, field
from typing import List


@dataclass
class UserInfo:
    userid: str = None
    area: str = None
    plat: str = None
    partition: str = None
    user_type: str = None
    role_id: str = None
    extra: dict = field(default_factory=dict)

    def to_dict(self):
        return asdict(self)

    def parse(self, info_dict):
        for key, value in info_dict.items():
            setattr(self, key, value)


class UserInfoGenerator:
    """ 用户信息生成器 """
    def __init__(self, field_list: list[str]):
        self.field_list = field_list

    def generate(self) -> List[UserInfo]:
        raise NotImplementedError



import ast
import inspect
import sys
from typing import List, Callable

import numpy as np

from .session import Session, HttpTransaction
from .testcase import SingleSessionCase, SingleRequestCase, AllSessionCase, TestCase

_session_checker_prefix = "chk"


def default_session_checker_prefix():
    return _session_checker_prefix


def _dist_dict_to_list(dist: dict, format_ratio=True):
    kv_list = list(dist.items())
    kv_list = sorted(kv_list, key=lambda x: x[1], reverse=True)
    output = []
    total = 0
    for _, v in kv_list:
        total += v

    for k, v in kv_list:
        ratio = v / total
        if format_ratio:
            output.append({"group": k, "count": v, "ratio": f"{ratio * 100:.2f}%"})
        else:
            output.append({"group": k, "count": v, "ratio": ratio})
    output = sorted(output, key=lambda x: (-x["count"], x["group"]))
    return output


def transaction_elem_dist_stat(session_list: List[Session], custom_flag_func: Callable, format_ratio=True):
    """HTTP transaction级别元素分布统计
    """
    dist = {}
    for ss in session_list:
        for s in ss.transactions:
            rsp = s.rsp_json()
            flag = custom_flag_func(rsp)
            if isinstance(flag, list):
                for f in flag:
                    dist[f] = dist.get(f, 0) + 1
            else:
                dist[flag] = dist.get(flag, 0) + 1
    return _dist_dict_to_list(dist, format_ratio)


def session_elem_dist_stat(session_list: List[Session], custom_flag_func: Callable, format_ratio=True):
    """session级别元素分布统计
    """
    dist = {}
    for s in session_list:
        flag = custom_flag_func(s)
        if isinstance(flag, list):
            for f in flag:
                dist[f] = dist.get(f, 0) + 1
        else:
            dist[flag] = dist.get(flag, 0) + 1
    return _dist_dict_to_list(dist, format_ratio)


def stat_http_transaction_cost(session_list: List[Session]):
    """统计请求耗时，按照平均值，中位值，P90，P99进行统计"""
    request_times = []
    for s in session_list:
        for t in s.transactions:
            if t.cost_time is not None:
                request_times.append(t.cost_time)

    # 计算平均值
    mean_time = np.mean(request_times)

    # 计算中位值
    median_time = np.median(request_times)

    # 计算 P90
    p90_time = np.percentile(request_times, 90)

    # 计算 P99
    p99_time = np.percentile(request_times, 99)

    report = [
        {"耗时类型": "平均值", "耗时": f"{int(mean_time * 100)}ms"},
        {"耗时类型": "P50", "耗时": f"{int(mean_time * 100)}ms"},
        {"耗时类型": "P90", "耗时": f"{int(p90_time * 100)}ms"},
        {"耗时类型": "P99", "耗时": f"{int(p99_time * 100)}ms"},
    ]

    return (mean_time, median_time, p90_time, p99_time), report


def func_to_case(name: str, func) -> TestCase:
    signature = inspect.signature(func)
    params = list(signature.parameters.values())
    if params:
        if params[0].annotation == HttpTransaction:
            # 单请求
            case = SingleRequestCase(rsp_checker=func)
        elif params[0].annotation == Session:
            # 单会话
            case = SingleSessionCase(session_checker=func)
        elif params[0].annotation == List[Session]:
            # 所有会话
            case = AllSessionCase(session_list_checker=func)
        else:
            raise ValueError(
                f"First parameter of function {name} should be of type HttpTransaction or Session")

        if not case.name:
            raise ValueError(f"Function {name} should have a name")
        if not case.expectation:
            raise ValueError(f"Function {name} should have an expectation")
        return case

    raise ValueError(f"Function {name} should have at least one parameter")


def auto_gen_cases_from_chk_func(checker_prefix=_session_checker_prefix, module_name="__main__"):
    module = sys.modules[module_name]
    source = inspect.getsource(module)
    tree = ast.parse(source)

    check_cases = []

    for node in ast.walk(tree):
        if isinstance(node, ast.FunctionDef) and node.name.startswith(checker_prefix):
            func = getattr(module, node.name)
            case = func_to_case(node.name, func)
            check_cases.append(case)

    return check_cases



import uuid
from typing import List

from demo.demo_svr import calc_sig
from session_tester import Session, UserInfo, Tester, TestSuite, CheckResult, HttpTransaction, SessionMaintainerBase, \
    ts_with_http_cost_stat


class SessionMaintainer(SessionMaintainerBase):
    def __init__(self):
        super().__init__("http://localhost:8000", http_method="POST")

    def load_user_info(self):
        for i in range(10):
            self.user_info_queue.put(UserInfo(userid=uuid.uuid4().hex, extra={"index": i}))

    @staticmethod
    def init_session(s: Session):
        s.ext_state.update({"items": [], "round": 0})

    @staticmethod
    def wrap_req(s: Session):
        ui = s.user_info
        items_owned = s.ext_state.get("items", [])
        round_ = s.ext_state.get("round", 0)
        return {"user_id": ui.userid, "round": round_, "items_owned": items_owned}

    @staticmethod
    def should_stop_session(s: Session) -> bool:
        return len(s.transactions) >= 20

    @staticmethod
    def update_session(s: Session):
        o = s.transactions[-1].rsp_json()
        s.ext_state["items"] += o["items"]
        s.ext_state["round"] = o["next_round"]


@ts_with_http_cost_stat
class T(TestSuite):
    """测试模块"""

    @staticmethod
    def chk_rsp_sig(s: HttpTransaction) -> CheckResult:
        """单请求-签名校验:
        1. 接口返回sig字段
        2. sig字段与user_id, round, items计算结果一致
        """
        req = s.req_json()
        rsp = s.rsp_json()
        sig = calc_sig(rsp.get("user_id", ""), req.get("round", 0), rsp.get("items", []))
        if sig == rsp.get("signature"):
            return CheckResult(True, "sig check pass")
        return CheckResult(False, f"sig check failed, {s.response}")

    @staticmethod
    def chk_no_repeat_item(s: HttpTransaction) -> CheckResult:
        """单请求-返回道具重复性校验:
        1. 返回结果中items字段不重复
        """
        items = s.rsp_json().get("items", [])
        if len(set(items)) == len(items):
            return CheckResult(True)
        return CheckResult(False, f"items repeat, {s.response}")

    @staticmethod
    def chk_no_repeat_item_in_all_sessions(s: Session) -> CheckResult:
        """
        单用户-返回道具重复性校验:
        1. 所有请求返回结果中items字段不重复
        """
        item_set = set()
        for t in s.transactions:
            items = t.rsp_json().get("items", [])
            for item in items:
                if item in item_set:
                    return CheckResult(False, f"item[{item}] repeat, {t}")
                item_set.add(item)
        return CheckResult(True)

    @staticmethod
    def chk_items_dist_in_all_sessions(ss: List[Session]) -> CheckResult:
        """
        返回道具分布校验:
        1. 所有请求返回结果中items字段分布均匀
        """
        # stat_result = transaction_elem_dist_stat(ss, lambda x: x.get("items", []))
        # return CheckResult(True, "", stat_result)
        # 以下代码可以通过上边两行代替

        item_dist = {}
        for s in ss:
            for t in s.transactions:
                items = t.rsp_json().get("items", [])
                for item in items:
                    item_dist[item] = item_dist.get(item, 0) + 1

        stat_result = [{"item": k, "count": v} for k, v in sorted(item_dist.items())]
        return CheckResult(True, "", stat_result)


def main():
    # 1. 创建测试对象
    t = Tester(
        name=f"release_12345",
        test_suites=[
            T(session_maintainer=SessionMaintainer()),
            T(session_maintainer=SessionMaintainer(), name="测试模块2"),
        ],
    )

    # 2. 运行测试用例产生报告
    t.run(
        mode=Tester.RUN_MODE_NEW,
    )


if __name__ == "__main__":
    main()



import glob
import json
import math
import os
import threading
from dataclasses import asdict
from dataclasses import dataclass
from datetime import datetime
from typing import List, Optional

from .logger import logger
from .user_info import UserInfo

test_session_dir = os.getenv("TEST_SESSION_DIR", "./test_sessions")
if not os.path.exists(test_session_dir):
    os.makedirs(test_session_dir)
sub_session_dir_exists = False


def update_test_session_dir(name: str):
    global test_session_dir, sub_session_dir_exists
    test_session_dir = os.path.join(test_session_dir, name)
    if not os.path.exists(test_session_dir):
        os.makedirs(test_session_dir)

    sub_session_dir_exists = True


@dataclass
class HttpTransaction:
    url: str  # 存储请求的URL
    method: str  # 存储请求的方法
    status_code: Optional[int]  # 存储HTTP状态码
    request: Optional[str]  # 存储请求数据（序列化后的字符串）
    response: Optional[str]  # 存储响应数据（序列化后的字符串）
    request_time: Optional[datetime] = datetime.now()  # 存储请求时间
    cost_time: Optional[float] = 0.0
    retry_cnt: Optional[int] = 0

    def to_json(self) -> str:
        return json.dumps(self.to_dict(), indent=2)

    def to_dict(self) -> dict:
        # 将 datetime 对象转换为字符串
        data = asdict(self)
        data['request_time'] = self.request_time.isoformat()
        return data

    def req_json(self):
        return json.loads(self.request)

    def rsp_json(self):
        return json.loads(self.response)

    def rsp_json_data(self):
        return self.rsp_json()["data"]

    def finished_without_error(self):
        return self.status_code == 200

    @staticmethod
    def from_json(json_str: str) -> 'HttpTransaction':
        data = json.loads(json_str)
        # 将字符串转换回 datetime 对象
        data['request_time'] = datetime.fromisoformat(data['request_time'])
        return HttpTransaction(**data)


class IDGenerator:
    _id = None
    _lock = threading.Lock()  # 使用线程锁来保护类变量
    id_dict = {}

    @classmethod
    def _read_initial_id(cls, file_path):
        k = file_path
        file_path = os.path.join(test_session_dir, file_path)
        try:
            with open(file_path, 'r') as file:
                cls.id_dict[k] = int(file.read().strip())
        except:
            cls.id_dict[k] = 0

    @classmethod
    def _write_id_to_file(cls, file_path: str):
        k = file_path
        file_path = os.path.join(test_session_dir, file_path)
        with open(file_path, 'w') as file:
            file.write(str(cls.id_dict[k]))

    @classmethod
    def get_next_id(cls, file_path: str) -> int:
        with cls._lock:  # 使用线程锁来保护临界区
            if cls.id_dict.get(file_path, None) is None:
                cls._read_initial_id(file_path)
            cls.id_dict[file_path] += 1
            cls._write_id_to_file(file_path)
            return cls.id_dict[file_path]

    @classmethod
    def get_curr_id(cls, file_path: str) -> int:
        with cls._lock:  # 使用线程锁来保护临界区
            cls._read_initial_id(file_path)
            return cls.id_dict[file_path]


class Session(IDGenerator):

    def __init__(self, label: str, create_flag=True):
        self.label = label
        self.user_info: UserInfo = None
        self.transactions: List[HttpTransaction] = []
        self.start_time = None
        self.session_filename = None
        self.no_dump = False
        self.ext_state = {}
        if create_flag:
            self.session_id = Session.get_next_id(label)

    def finished_without_error(self):
        return all([x.finished_without_error() for x in self.transactions])

    @staticmethod
    def load_session(session_filename: str) -> 'Session':
        with open(session_filename, 'r') as file:
            return Session.from_json(file.read())

    @staticmethod
    def clear_sessions(label: str):
        if sub_session_dir_exists:
            # remote all files in test_session_dir but not remove the dir
            for root, _, files in os.walk(test_session_dir, topdown=False):
                for name in files:
                    os.remove(os.path.join(root, name))
            return
        id_file = os.path.join(test_session_dir, f"{label}")
        session_filename_list = glob.glob(id_file + "-*.json")
        files = session_filename_list + [id_file]
        for filename in files:
            try:
                os.remove(filename)
            except Exception as e:
                logger.error("Failed to remove session {%s}: {%s}", filename, e)

    @staticmethod
    def load_sessions(label: str, n: int = math.inf) -> List['Session']:
        sessions = []
        id_ = Session.get_curr_id(label)
        while id_ > 0 and len(sessions) < n:
            session_filename = f"{label}-{id_:08d}.json"
            # TODO 数据量过大时，可以考虑分批加载和校验
            try:
                s = Session.load_session(os.path.join(test_session_dir, session_filename))
                sessions.append(s)
            except Exception as e:
                logger.error("Failed to load session {%s}: {%s}", session_filename, e)
            id_ -= 1
        return sessions

    def create(self, user_info: UserInfo, transactions: List[HttpTransaction],
               start_time: Optional[float] = None,
               no_dump: bool = False) -> 'Session':
        self.user_info = user_info
        self.transactions = transactions
        self.start_time = start_time
        # 创建一个session文件
        self.session_filename = f"{self.label}-{self.session_id:08d}.json"
        self.ext_state = {}
        self.no_dump = no_dump
        if not no_dump:
            self.dump()
        return self

    def append_transaction(self, transaction: HttpTransaction):
        self.transactions.append(transaction)

    def to_json(self) -> str:
        return json.dumps({
            'label': self.label,
            'session_id': self.session_id,
            'user_info': self.user_info.to_dict() if self.user_info else None,
            'transactions': [x.to_dict() for x in self.transactions],
            'ext_state': self.ext_state,
            'start_time': self.start_time
        }, indent=2)

    def dump(self):
        if self.session_filename:
            if self.no_dump:
                return
            full_session_filename = os.path.join(test_session_dir, self.session_filename)
            with open(full_session_filename, 'w') as file:
                file.write(self.to_json())
        else:
            raise ValueError("Session filename is not set")

    @staticmethod
    def from_json(json_str: str) -> 'Session':
        data = json.loads(json_str)
        user_info = UserInfo(**data['user_info'])  # 假设 UserInfo 类可以通过 **kwargs 初始化
        if not data['transactions']:
            raise ValueError("No transactions found")

        transactions = [HttpTransaction(**tx) for tx in data['transactions']]
        s = Session(label=data['label'], create_flag=False)
        s.user_info = user_info
        s.transactions = transactions
        s.start_time = data.get('start_time')
        s.session_id = data['session_id']
        s.ext_state = data['ext_state']
        s.session_filename = f"{s.label}-{s.session_id:08d}.json"
        return s



from .client import Client
from .decorator import SessionMaintainerSimple, sm_n_rounds, sm_no_update, sm_no_init, sm_simple_n, \
    ts_with_http_cost_stat
from .session import Session, HttpTransaction
from .session_maintainer import SessionMaintainerBase
from .test_suite import TestSuite
from .testcase import SingleRequestCase, SingleSessionCase, AllSessionCase, CheckResult
from .tester import Tester
from .user_info import UserInfo
from .utils import auto_gen_cases_from_chk_func

__all__ = ["Client", "Session", "UserInfo", "SingleRequestCase", "SingleSessionCase", "AllSessionCase", "Tester",
           "auto_gen_cases_from_chk_func", "TestSuite", "CheckResult", "HttpTransaction", "SessionMaintainerBase",
           "SessionMaintainerSimple",
           # session_maintainer.py
           "sm_n_rounds", "sm_no_update", "sm_no_init", "sm_simple_n",
           # test_suite decorators
           "ts_with_http_cost_stat",
           ]



import hashlib
import json
import random
from http.server import BaseHTTPRequestHandler, HTTPServer


def calc_sig(user_id, round_num, items):
    return hashlib.md5((user_id + str(round_num) + ",".join([str(x) for x in items])).encode('utf-8')).hexdigest()


class SimpleHTTPRequestHandler(BaseHTTPRequestHandler):
    def _set_response(self):
        self.send_response(200)
        self.send_header('Content-type', 'application/json')
        self.end_headers()

    def do_POST(self):
        content_length = int(self.headers['Content-Length'])
        post_data = self.rfile.read(content_length)
        data = json.loads(post_data)
        user_id = data.get('user_id')
        round_num = data.get('round')
        items_owned = data.get('items_owned', [])
        candidate_items = set(list(range(100))) - set(items_owned)
        # items = random.choices(list(candidate_items), k=5)
        items = random.sample(list(candidate_items), k=5)
        sig = calc_sig(user_id, round_num, items)
        response_data = {
            "user_id": user_id,
            "next_round": round_num + 1,
            "items": list(items),
            "signature": sig,
        }
        self._set_response()
        self.wfile.write(json.dumps(response_data).encode('utf-8'))


def run(server_class=HTTPServer, handler_class=SimpleHTTPRequestHandler, port=8000):
    server_address = ('', port)
    httpd = server_class(server_address, handler_class)
    print(f'Starting httpd on port {port}')
    httpd.serve_forever()


if __name__ == '__main__':
    run()



import ast
import datetime
import inspect
import queue
import threading
import time
from dataclasses import dataclass
from typing import List

from .client import Client
from .logger import logger
from .session import Session
from .session_maintainer import SessionMaintainerBase
from .testcase import TestCase, SingleRequestCase, Report, SingleSessionCase, AllSessionCase
from .utils import func_to_case, default_session_checker_prefix


class TestSuite:
    def __init__(self, name=None, session_maintainer: SessionMaintainerBase = None, spec_cases=None):
        self.name = name
        if name is None:
            self.name = self.__doc__

        if not self.name:
            raise RuntimeError("test suite name is required")
        self.name = self.name.strip()

        self.session_maintainer: SessionMaintainerBase = session_maintainer
        self.report_list: List[Report] = []
        self.parent_name = None
        self._check_cases = self.merge_cases(spec_cases)

    def merge_cases(self, spec_cases):
        if spec_cases is None:
            spec_cases = []

        ret = []
        inserted_cases = set()
        for case in spec_cases:
            if case.name in inserted_cases:
                raise ValueError(f"Duplicate case name: {case.name}")
            inserted_cases.add(case.name)
            ret.append(case)

        for case in self.auto_gen_test_cases():
            if case.name in inserted_cases:
                raise ValueError(f"Duplicate case name: {case.name}, please change the name in chk func")
            inserted_cases.add(case.name)
            ret.append(case)
        return ret

    def check_cases(self):
        return self._check_cases

    @classmethod
    def auto_gen_test_cases(cls, inserted_check_func=None) -> List[TestCase]:
        if inserted_check_func is None:
            inserted_check_func = set()
        methods = [func for func in dir(cls) if callable(getattr(cls, func))]
        check_cases = []

        source = inspect.getsource(cls)
        tree = ast.parse(source)
        if len(tree.body) != 1 or not isinstance(tree.body[0], ast.ClassDef):
            raise ValueError("Only one class is allowed in the suite")
        tree = tree.body[0]

        # 有序的检查
        for node in tree.body:
            if isinstance(node, ast.FunctionDef) and node.name.startswith(default_session_checker_prefix()):
                func = getattr(cls, node.name)
                if isinstance(func, ast.FunctionDef):
                    func = getattr(cls, func.name)
                func_name = node.name
                if isinstance(cls.__dict__.get(func.__name__), staticmethod) and func_name not in inserted_check_func:
                    inserted_check_func.add(func_name)
                    case = func_to_case(func_name, func)
                    check_cases.append(case)

        # 处理修饰器的检查
        for method_name in methods:
            method = getattr(cls, method_name)
            # print(f"Method: {method}")
            if inspect.isfunction(method) or inspect.ismethod(method):
                if isinstance(cls.__dict__.get(method_name), staticmethod) and method_name.startswith(
                        default_session_checker_prefix()):
                    if method_name in inserted_check_func:
                        continue

                    inserted_check_func.add(method_name)
                    func = cls.__dict__.get(method_name)
                    case = func_to_case(method_name, func)
                    check_cases.append(case)
                    # sig = inspect.signature(method)
        # 处理父类中的函数
        for base in inspect.getmro(cls):
            if base == cls or not isinstance(base, TestSuite):
                continue
            try:
                check_cases += base.auto_gen_test_cases(inserted_check_func)
            except AttributeError:
                pass

        return check_cases

    def do_send(self, thread_cnt=50, no_dump=False):
        stopped = threading.Event()
        stopped.clear()
        lock = threading.Lock()

        logger.info(f"{self.name} 开始发送")

        @dataclass
        class SendStat:
            total_session_cnt: int = 0
            total_session_cost: int = 0
            total_send_cnt: int = 0
            total_send_err_cnt: int = 0
            total_retry_cnt: int = 0
            total_send_cost: int = 0
            start_time: datetime.datetime = 0
            end_time: datetime.datetime = 0

            def report(self):
                logger.info("发送请求统计：")
                logger.info(f"    {self.total_session_cnt} 个会话")
                logger.info(
                    f"    {self.total_send_cnt} 个请求(失败重试 {self.total_retry_cnt}, 最终失败 {self.total_send_err_cnt})")
                logger.info(f"    总耗时: {((self.end_time - self.start_time).total_seconds()):.2f} 秒")
                logger.info(f"    请求平均耗时: {(self.total_send_cost * 1000 / self.total_send_cnt):.2f} 毫秒")
                logger.info(f"    会话平均耗时: {(self.total_session_cost * 1000 / self.total_session_cnt):.2f} 毫秒")
                logger.info(f"    QPS: {(self.total_send_cnt / (self.end_time - self.start_time).total_seconds()):.2f}")

        send_stat = SendStat()

        class SendWorker(threading.Thread):
            def __init__(self, label, session_maintainer_cls: SessionMaintainerBase):
                threading.Thread.__init__(self)
                self.label = label
                self.user_info_queue = session_maintainer_cls.user_info_queue
                self.session_maintainer_cls = session_maintainer_cls

            def run(self):
                while True:
                    try:
                        user_info = self.user_info_queue.get_nowait()
                        session = Session(label=self.label)
                        session.create(user_info=user_info, transactions=[], no_dump=no_dump)
                        client = Client(session=session, session_maintainer=self.session_maintainer_cls)

                        start_time = datetime.datetime.now()
                        client.run()
                        elapsed_time = (datetime.datetime.now() - start_time).total_seconds()  # 计算请求时间
                        session.dump()
                        with lock:
                            nonlocal send_stat
                            send_stat.total_session_cnt += 1
                            send_stat.total_send_cnt += len(session.transactions)
                            send_stat.total_send_err_cnt += len(
                                [x for x in session.transactions if not x.finished_without_error()])
                            send_stat.total_retry_cnt += sum([x.retry_cnt for x in session.transactions])
                            send_stat.total_send_cost += sum([x.cost_time for x in session.transactions])
                            send_stat.total_session_cost += elapsed_time
                    except queue.Empty:
                        if stopped.is_set():
                            return
                        time.sleep(0.1)

        class QueueLoader(threading.Thread):
            def __init__(self, session_maintainer_cls: SessionMaintainerBase):
                threading.Thread.__init__(self)
                self.user_info_queue = session_maintainer_cls.user_info_queue
                self.session_maintainer_cls = session_maintainer_cls

            def run(self):
                if not self.session_maintainer_cls.user_info_queue.empty():
                    stopped.set()
                    return

                self.session_maintainer_cls.load_user_info()
                stopped.set()

        t_list = [QueueLoader(self.session_maintainer)]

        q = self.session_maintainer.user_info_queue
        if not q.empty() and q.qsize() > 0:
            thread_cnt = min(thread_cnt, q.qsize())

        for _ in range(thread_cnt):
            t = SendWorker(self.name, self.session_maintainer)
            t_list.append(t)

        send_stat.start_time = datetime.datetime.now()
        for t in t_list:
            t.start()

        for t in t_list:
            t.join()
        send_stat.end_time = datetime.datetime.now()
        return send_stat

    def clear_sessions(self):
        Session.clear_sessions(self.name)

    def check(self):
        # 加载会话结果
        session_list = Session.load_sessions(self.name)
        self.report_list = []
        for case in self.check_cases():
            if isinstance(case, SingleRequestCase):
                report = Report(case.name, case.expectation, "SingleRequestCase")
                transactions = [transaction for session in session_list for transaction in session.transactions]
                report.total_case_count = len(transactions)
                report.finished_with_err_count = len([x for x in transactions if not x.finished_without_error()])
                report.case_results = case.batch_check([x for x in transactions if x.finished_without_error()])

            elif isinstance(case, SingleSessionCase):
                report = Report(case.name, case.expectation, "SingleSessionCase")
                report.finished_with_err_count = len([x for x in session_list if not x.finished_without_error()])
                report.total_case_count = len(session_list)
                report.case_results = case.batch_check([x for x in session_list if x.finished_without_error()])
            elif isinstance(case, AllSessionCase):
                report = Report(case.name, case.expectation, "AllSessionCase")
                session_list = [x for x in session_list if x.finished_without_error()]
                if not session_list:
                    report.uncover_case_count = 1
                else:
                    report.case_results = case.batch_check([session_list])
            else:
                raise RuntimeError("unknown case type")

            for result in report.case_results:
                if result is None:
                    report.uncover_case_count += 1
                elif result.result:
                    report.passed_case_count += 1
                else:
                    report.not_passed_case_count += 1

            self.report_list.append(report)
            logger.info(f"{self.name}-{case.name} 检查完成")

        return self.report_list



from dataclasses import dataclass
from typing import Optional, Any


@dataclass
class StReq:
    req_data: Any
    url: Optional[str] = None
    http_method: Optional[str] = None
    timeout: Optional[tuple] = (1, 5)
    headers: Optional[dict] = None
    retry: Optional[int] = 1



from typing import List

from .session import Session
from .session_maintainer import SessionMaintainerBase
from .testcase import CheckResult
from .utils import stat_http_transaction_cost


# 为测试套件添加请求耗时统计
def ts_with_http_cost_stat(cls):
    @staticmethod
    def chk_http_cost_dist(ssl: List[Session]):
        """检查请求耗时分布:
        None
        """
        _, report = stat_http_transaction_cost(ssl)
        return CheckResult(True, None, report)

    cls.chk_http_cost_dist = chk_http_cost_dist
    return cls


# 以下是SessionMaintainer的装饰器

def sm_n_rounds(n: int):
    """ N 轮后停止 """

    def decorator(cls):
        @staticmethod
        def should_stop_session(s: Session) -> bool:
            return len(s.transactions) >= n

        cls.should_stop_session = should_stop_session
        return cls

    return decorator


def sm_no_update(cls):
    """ 不更新会话状态 """

    @staticmethod
    def update_session(_: Session):
        pass

    cls.update_session = update_session
    return cls


def sm_no_init(cls):
    """ 默认初始化状态 """

    @staticmethod
    def init_session(_: Session):
        pass

    cls.init_session = init_session
    return cls


def sm_simple_n(n: int):
    """ 简单N轮停止 """

    def decorator(cls):
        sm_no_init(cls)
        sm_no_update(cls)
        sm_n_rounds(n)(cls)

        @staticmethod
        def should_stop_session(s: Session) -> bool:
            return len(s.transactions) >= n

        cls.should_stop_session = should_stop_session
        return cls

    return decorator


@sm_simple_n(1)
class SessionMaintainerSimple(SessionMaintainerBase):
    """ 单轮会话维持 """
    pass



import logging

logger = logging.getLogger('session-tester')
logger.setLevel(logging.DEBUG)

console_handler = logging.StreamHandler()
console_handler.setLevel(logging.DEBUG)
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
console_handler.setFormatter(formatter)
logger.addHandler(console_handler)



import os
from typing import List

import pandas as pd
from openpyxl.reader.excel import load_workbook
from openpyxl.styles import Alignment, Font

from .logger import logger
from .session import update_test_session_dir
from .test_suite import TestSuite
from .testcase import Report

test_report_dir = os.getenv("TEST_REPORT_DIR", "./test_reports")
if not os.path.exists(test_report_dir):
    os.makedirs(test_report_dir)


class Tester:
    RUN_MODE_NEW = 0
    RUN_MODE_CHECK = 1
    RUN_MODE_BENCHMARK = 2

    def __init__(self,
                 name: str,
                 test_suites: List[TestSuite]):
        self.name = name
        self.test_suites = test_suites
        # Check no duplicate names
        names = [ts.name for ts in test_suites]
        if len(names) != len(set(names)):
            raise ValueError("Duplicate test suite names")

        for test_suite in test_suites:
            names = [tc.name for tc in test_suite.check_cases()]
            if len(names) != len(set(names)):
                raise ValueError(f"Duplicate test case names in suite {test_suite.name}")
        update_test_session_dir(self.name)

    def run(self, mode=RUN_MODE_NEW, thread_cnt=50):
        if mode not in [self.RUN_MODE_NEW, self.RUN_MODE_CHECK, self.RUN_MODE_BENCHMARK]:
            raise ValueError(f"Invalid tester run mode: {mode}")

        if mode == Tester.RUN_MODE_NEW:
            for test_suite in self.test_suites:
                test_suite.clear_sessions()
            logger.info("清除会话数据成功")
            for test_suite in self.test_suites:
                test_suite.do_send(thread_cnt=thread_cnt)
            logger.info("发送请求完成")
        elif mode == Tester.RUN_MODE_BENCHMARK:
            logger.info("启动压力测试")
            for test_suite in self.test_suites:
                result = test_suite.do_send(thread_cnt=thread_cnt, no_dump=True)
                result.report()
            logger.info("压测请求完成")

        # 只有新模式和校验模式下才会执行校验
        if mode in [Tester.RUN_MODE_NEW, Tester.RUN_MODE_CHECK]:
            for test_suite in self.test_suites:
                test_suite.check()
                logger.info(f"{test_suite.name}校验完成")

            # 保存为Excel文件
            output_file = os.path.join(test_report_dir, f"测试报告-{self.name}.xlsx")
            with pd.ExcelWriter(output_file, engine='xlsxwriter') as writer:
                self.gen_summary(writer)
                self.gen_detail_report(writer)
            self.format()

    def gen_summary(self, writer):
        # 解析数据
        parsed_data = []
        reports: List[Report]
        for test_suite in self.test_suites:
            reports = test_suite.report_list
            for report in reports:
                report.summary()
                # 创建一个包含字典数据的列表
                parsed_data.append({
                    "功能模块": test_suite.name,
                    "功能点": report.name,
                    "预期结果": report.expectation,
                    "最终结果": report.result,
                    "通过": report.passed_case_count,
                    "未通过": report.not_passed_case_count,
                    "未覆盖": report.uncover_case_count,
                    "网络错误": report.finished_with_err_count,
                    "异常说明": report.bad_case
                })

            # 保存为Excel文件
            output_file = os.path.join(test_report_dir, f"测试报告-{self.name}.xlsx")

            # 汇总信息
            df = pd.DataFrame(parsed_data)
            df.to_excel(writer, sheet_name="测试汇总", index=False)

        logger.info(f"汇总报告-已成功保存到 {output_file}")

    def gen_detail_report(self, writer):

        k = set()
        dup_test_case_name_set = set()
        for test_suite in self.test_suites:
            reports = test_suite.report_list
            for report in reports:
                report.summary()
                if not report.ext_report:
                    continue
                if report.name in k:
                    dup_test_case_name_set.add(report.name)
                k.add(report.name)

        for test_suite in self.test_suites:
            reports = test_suite.report_list
            for report in reports:
                report.summary()
                if not report.ext_report:
                    continue

                # 保存为Excel文件
                ext_df = pd.DataFrame(report.ext_report)
                if report.name in dup_test_case_name_set:
                    sheet_name = f"{report.name}({test_suite.name})"
                else:
                    sheet_name = report.name

                ext_df.to_excel(writer, sheet_name=sheet_name, index=False)
                logger.info(f"详细数据-已成功保存到 表-{sheet_name}")

    def format(self):
        output_file = os.path.join(test_report_dir, f"测试报告-{self.name}.xlsx")
        # 加载生成的Excel文件
        wb = load_workbook(output_file)
        ws = wb.active

        # 设置列宽
        column_widths = {
            'A': 30,  # 功能模块
            'B': 40,  # 功能点
            'C': 60,  # 预期结果
            'D': 15,  # 最终结果
            'E': 8.5,  # 通过
            'F': 10,  # 未通过
            'G': 10,  # 未覆盖
            'H': 13,  # 网络错误
            'I': 55,  # 异常说明
        }
        for col, width in column_widths.items():
            ws.column_dimensions[col].width = width

        font = Font(size=16)
        for row in ws.iter_rows():
            for cell in row:
                cell.font = font
                if cell.row == 1:
                    cell.alignment = Alignment(horizontal='center', vertical='center', wrap_text=True)
                    cell.font = Font(size=16, bold=True)
                elif cell.column_letter in ['C', 'I']:
                    cell.alignment = Alignment(horizontal='left', vertical='center', wrap_text=True)
                else:  # elif cell.column_letter in ['A', 'B', 'D', 'E', 'G']:
                    cell.alignment = Alignment(horizontal='center', vertical='center', wrap_text=True)

        # 合并A列连续相同的单元格
        start_row = 2
        end_row = ws.max_row
        current_value = None
        merge_start = start_row
        for row in range(start_row, end_row + 1):
            cell_value = ws[f'A{row}'].value
            if cell_value != current_value:
                if merge_start < row - 1:
                    ws.merge_cells(start_row=merge_start, start_column=1, end_row=row - 1, end_column=1)
                current_value = cell_value
                merge_start = row

        # 最后一组相同值的单元格合并
        if merge_start < end_row:
            ws.merge_cells(start_row=merge_start, start_column=1, end_row=end_row, end_column=1)

        # 自动调整行高
        for row in ws.iter_rows(min_row=1, max_row=ws.max_row):
            max_height = 0
            for cell in row:
                if cell.value:
                    lines = str(cell.value).split('\n')
                    max_height = max(max_height, len(lines))
            if max_height > 0:
                ws.row_dimensions[row[0].row].height = max_height * 20

        # 保存调整后的Excel文件
        wb.save(output_file)



import queue

from .request import StReq
from .session import Session


class SessionMaintainerBase:
    url: str = None
    http_method: str = "POST"

    def __init__(self, url: str, http_method: str = "POST"):
        self.url = url
        self.http_method = http_method
        self.user_info_queue = queue.Queue()

    def load_user_info(self):
        if self.user_info_queue.empty():
            raise RuntimeError("user_info_queue is empty")

    @staticmethod
    def init_session(_: Session):
        raise NotImplementedError

    @staticmethod
    def wrap_req(_: Session) -> StReq:
        raise NotImplementedError

    @staticmethod
    def update_session(_: Session):
        raise NotImplementedError

    @staticmethod
    def should_stop_session(_: Session) -> bool:
        raise NotImplementedError



import traceback
from dataclasses import dataclass
from typing import Callable, List, Optional, Any

from .session import HttpTransaction, Session


@dataclass
class CheckResult:
    result: bool
    exception: Optional[str] = None
    report_lines: Optional[List[Any]] = None


class TestCase:
    def __init__(self, name: str, expectation: str):
        self.name = name
        self.expectation = expectation

    def check(self, _: Any) -> CheckResult:
        raise NotImplementedError

    def batch_check(self, arg_list: List[Any]) -> List[CheckResult]:
        ret = []
        for arg in arg_list:
            try:
                check_result = self.check(arg)
            except Exception as e:
                # 获取异常堆栈信息
                stack_trace = traceback.format_exc()
                check_result = CheckResult(False, f"checking exception: {e}\nStack trace:\n{stack_trace}", None)

            ret.append(check_result)
        return ret


def overwrite_name_and_expectation(name, expectation, doc):
    if doc is None:
        name_, expectation_ = None, None
    else:
        if doc.find(":") == -1:
            # 避免用户使用中文冒号
            doc = doc.replace("：", ":")
        fields = doc.split(":", 1)
        name_ = fields[0].strip()
        if len(fields) > 1:
            expectation_ = '\n'.join([x.lstrip(" ") for x in fields[1].strip().split("\n")])
        else:
            expectation_ = None
    if name is None:
        name = name_
    if expectation is None:
        expectation = expectation_
    if not name:
        raise RuntimeError("name is required")

    return name, expectation


# 单个请求检查
class SingleRequestCase(TestCase):
    def __init__(self, name: str = None, expectation: str = None,
                 rsp_checker: Callable[[HttpTransaction], CheckResult] = None):
        name, expectation = overwrite_name_and_expectation(name, expectation, rsp_checker.__doc__)
        super().__init__(name, expectation)
        self.rsp_checker = rsp_checker

    def check(self, transaction: HttpTransaction) -> CheckResult:
        if self.rsp_checker is None:
            raise RuntimeError("rsp_checker is required")
        return self.rsp_checker(transaction)


# 单个会话检查
class SingleSessionCase(TestCase):
    def __init__(self, name: str = None, expectation: str = None,
                 session_checker: Callable[[Session], CheckResult] = None):
        name, expectation = overwrite_name_and_expectation(name, expectation, session_checker.__doc__)
        super().__init__(name, expectation)
        self.session_checker = session_checker

    def check(self, session: Session):
        if self.session_checker is None:
            raise RuntimeError("session_checker is required")
        return self.session_checker(session)


# 全体会话检查
class AllSessionCase(TestCase):
    def __init__(self, name: str = None, expectation: str = None,
                 session_list_checker: Callable[[List[Session]], CheckResult] = None):
        name, expectation = overwrite_name_and_expectation(name, expectation, session_list_checker.__doc__)
        super().__init__(name, expectation)
        self.session_list_checker = session_list_checker

    def check(self, session_list: List[Session]):
        if self.session_list_checker is None:
            raise RuntimeError("session_list_checker is required")
        return self.session_list_checker(session_list)


class Report:
    def __init__(self, name: str, expectation: str, case_type: str):
        self.name = name
        self.expectation = expectation
        self.case_type = case_type
        self.result = None
        self.bad_case = None
        self.ext_report = []
        self.case_results: List[CheckResult] = []
        self.finished_with_err_count = 0
        self.passed_case_count = 0
        self.not_passed_case_count = 0
        self.uncover_case_count = 0

    def summary(self):
        if not self.case_results:
            self.result = "未覆盖"
            return
        self.result = "通过"

        self.ext_report = []
        for case_result in self.case_results:
            if case_result is not None and case_result.report_lines is not None:
                self.ext_report += case_result.report_lines

        for case_result in self.case_results:
            if case_result is not None and not case_result.result:
                self.result = "未通过"
                self.bad_case = case_result.exception
                return

        if self.uncover_case_count > 0 and self.passed_case_count == 0:
            self.result = "未覆盖"

    def __str__(self):
        self.summary()
        return f"{self.name} {self.expectation} {self.case_type} {self.result} {self.ext_report}"
